#include "paramSimul.h"
#include "ue.h"
#include "objectBSSectorUe.h"
#include "paramFastFading.h"
#include "rvGenerater.h"
#include "evaluationConfigurerations.h"
#include <algorithm>


void Generate_channel_samples(int ue_idx, int adj_sec_idx);
void Generate_channel_coef_elements(int ue_idx, int sec_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double  GCS_theta_LOS_ZOD, int N, int M, int fast_fading_condition);
void Generate_delays(double K, double DS, double *tau,double *delay, int num_cluster, int fast_fading_condition);
int Generate_cluster_powers(int ue_idx, int bs_idx, double K_R, double P_1LOS, double DS, double *tau, double *delay,  int num_cluster, int fast_fading_condition);
void Generate_arrival_departure_angle(int ue_idx, int bs_idx, double K, double ASA, double ASD, double ZSA, double ZSD,double mu_offset_ZOD, int num_cluster, int num_ray_per_cluster, int fast_fading_condition, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD, double phi_LOS_AOA,double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD);
void Couple_rays(int ue_idx, int sec_idx, int N, int M, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD);
void Shuffle_weak_ray(int *rays);
void Shuffle_strong_ray(int *rays);

void Generate_XPR(int ue_idx, int bs_idx, double mu_XPR, double sigma_XPR, int num_cluster, int num_ray_per_cluster);
double max(double *A, int size);
double Transform_angle_minus_180_to_plus_180(double x);
double Transform_angle_0_to_plus_180(double x);
void Find_strong_cluster(int ue_idx, int sec_idx, int N);

void Generate_initial_random_phase(int ue_idx, int bs_idx,int M,int N);
void Generate_LOS_field(int ue_idx,int sec_idx, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double GCS_theta_LOS_ZOD);
void Generate_NLOS_field(int ue_idx, int bs_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD, int N, int M);
void Generate_LOS_wave_direction_vector(int ue_idx, int bs_idx, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double GCS_theta_LOS_ZOD);
void Generate_NLOS_wave_direction_vector(int ue_idx, int bs_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD, int N, int M);

void Generate_LOS_UE_velocity(int ue_idx, int bs_idx, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double GCS_theta_LOS_ZOD);
void Generate_NLOS_UE_velocity(int ue_idx, int bs_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD, int N, int M);



LOCATION Transform_angle_to_spheical_vector(double phi, double theta);

double Sector_Antenna_gain_in_dB(SPHERICAL_ANGLE LCS_angle);
double UE_Antenna_gain_in_dB(SPHERICAL_ANGLE LCS_angle);

SPHERICAL_ANGLE GCS_Angle_to_LCS_Angle(double alpha, double beta, double gamma, SPHERICAL_ANGLE GCS_angle);
ANTENNA_FIELD Antenna_field_in_LCS(double LCS_antenna_gain_in_dB, double zeta);
ANTENNA_FIELD Antenna_field_in_LCS(double LCS_antenna_gain_in_dB, double zeta);
ANTENNA_FIELD LCS_Antenna_field_to_GCS_antenna_pattern(double alpha, double beta, double gamma, SPHERICAL_ANGLE GCS_angle, ANTENNA_FIELD LCS_field);

double dot(LOCATION a, LOCATION b);

void Generate_SSP()
{
	for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
	{

		cout << '\r';
		cout << int(double(ue_idx+1) / double(num_ue) * 100) << "%";
	
		for (int bs_idx = 0; bs_idx < num_bs; bs_idx++)
		{
			Generate_channel_samples(ue_idx, bs_idx);
		}
	}

}


/*===================================================================
FUNCTION: Generate_channel_samples()

AUTHOR:

DESCRIPTION: compute channel coefficients and stored

NOTES: Output is a Channel_Coef.
 
===================================================================*/
void Generate_channel_samples(int ue_idx, int bs_idx)
{
	
	int fast_fading_condition = channel[bs_idx][ue_idx].fast_fading_condition;
	int propagation_condition = channel[bs_idx][ue_idx].propagation_condition;
	int N = num_cluster[fast_fading_condition];
	int M = num_ray_per_cluster[fast_fading_condition];
	
		

	double GCS_phi_LOS_AOA = channel[bs_idx][ue_idx].GCS_Arrival_angle_LOS.phi;
	double GCS_phi_LOS_AOD = channel[bs_idx][ue_idx].GCS_Departure_angle_LOS.phi;
	double GCS_theta_LOS_ZOA = channel[bs_idx][ue_idx].GCS_Arrival_angle_LOS.theta;
	double GCS_theta_LOS_ZOD = channel[bs_idx][ue_idx].GCS_Departure_angle_LOS.theta;
	double mu_offset_ZOD = channel[bs_idx][ue_idx].ZOD_offset;

	double DS;
	double K = 0;
	double K_R = 0;
	double P_1LOS = 0;
	double *tau;
	double ASA;
	double ASD;
	double ZSA;
	double ZSD;
	double mu_XPR;
	double sigma_XPR;
	double **GCS_phi_AOA;
	double **GCS_phi_AOD;
	double **GCS_theta_ZOA;
	double **GCS_theta_ZOD;



	tau = new double[N];
		
	GCS_phi_AOA = new double *[N];
	GCS_phi_AOD = new double *[N];
	GCS_theta_ZOA = new double *[N];
	GCS_theta_ZOD = new double *[N];



	for (int n = 0; n < N; n++)
	{
		GCS_phi_AOA[n] = new double[M];
		GCS_phi_AOD[n] = new double[M];
		GCS_theta_ZOA[n] = new double[M];
		GCS_theta_ZOD[n] = new double[M];

	}


	
	switch (fast_fading_condition)
	{
	case LOS:
		DS = channel[bs_idx][ue_idx].LSPs[LOS_DS];
		K = channel[bs_idx][ue_idx].LSPs[LOS_K];
		ASA = channel[bs_idx][ue_idx].LSPs[LOS_ASA];
		ASD = channel[bs_idx][ue_idx].LSPs[LOS_ASD];
		ZSA = channel[bs_idx][ue_idx].LSPs[LOS_ZSA];
		ZSD = channel[bs_idx][ue_idx].LSPs[LOS_ZSD];
		K_R = pow(10, K / 10);
		P_1LOS = K_R/(K_R+1);
		mu_XPR = XPR_mean[LOS];
		sigma_XPR = XRP_std[LOS];

		break;
	case NLOS:
		DS = channel[bs_idx][ue_idx].LSPs[NLOS_DS];
		ASA = channel[bs_idx][ue_idx].LSPs[NLOS_ASA];
		ASD = channel[bs_idx][ue_idx].LSPs[NLOS_ASD];
		ZSA = channel[bs_idx][ue_idx].LSPs[NLOS_ZSA];
		ZSD = channel[bs_idx][ue_idx].LSPs[NLOS_ZSD];
		mu_XPR = XPR_mean[NLOS];
		sigma_XPR = XRP_std[NLOS];
		break;
	case OUT2IN:
		DS = channel[bs_idx][ue_idx].LSPs[OUT2IN_DS];
		ASA = channel[bs_idx][ue_idx].LSPs[OUT2IN_ASA];
		ASD = channel[bs_idx][ue_idx].LSPs[OUT2IN_ASD];
		ZSA = channel[bs_idx][ue_idx].LSPs[OUT2IN_ZSA];
		ZSD = channel[bs_idx][ue_idx].LSPs[OUT2IN_ZSD];
		mu_XPR = XPR_mean[OUT2IN];
		sigma_XPR = XRP_std[OUT2IN];
		break;
	}

	
	Generate_delays(K, DS, tau, channel[bs_idx][ue_idx].tau, N, fast_fading_condition);

	// tau[n]  channel[bs_idx][ue_idx].tau

	channel[bs_idx][ue_idx].num_cluster = Generate_cluster_powers(ue_idx, bs_idx, K_R, P_1LOS, DS,tau, channel[bs_idx][ue_idx].tau, N, fast_fading_condition);
	//channel[bs_idx][ue_idx].delay[n], channel[bs_idx][ue_idx].cluster_power[n] N개


	Generate_arrival_departure_angle(ue_idx, bs_idx, K,  ASA, ASD, ZSA, ZSD, mu_offset_ZOD, N, M, fast_fading_condition, GCS_phi_AOA, GCS_phi_AOD, GCS_theta_ZOA, GCS_theta_ZOD, GCS_phi_LOS_AOA, GCS_phi_LOS_AOD, GCS_theta_LOS_ZOA, GCS_theta_LOS_ZOD);
	//phi_AOD[n][m] ,	theta_ZOD[n][m],	phi_AOA[n][m], 	theta_ZOA[n][m]

	Couple_rays(ue_idx, bs_idx, N,M, GCS_phi_AOA, GCS_phi_AOD, GCS_theta_ZOA, GCS_theta_ZOD);

	//phi_AOA[cluster_idx][ray_idx] 	phi_AOD[cluster_idx][ray_idx] 	theta_ZOA[cluster_idx][ray_idx]	theta_ZOD[cluster_idx][ray_idx]
	//channel[bs_idx][ue_idx].strong_cluster[0]) channel[bs_idx][ue_idx].strong_cluster[1]

	Generate_XPR( ue_idx,  bs_idx, mu_XPR,sigma_XPR, N, M);
	// channel[bs_idx][ue_idx].kappa[n][m] = pow(10., (sigma_XPR*Rand.n() + mu_XPR) / 10.)

	Generate_channel_coef_elements(ue_idx, bs_idx, GCS_phi_AOA, GCS_phi_AOD, GCS_theta_ZOA, GCS_theta_ZOD, GCS_phi_LOS_AOA, GCS_phi_LOS_AOD, GCS_theta_LOS_ZOA, GCS_theta_LOS_ZOD, N, M, fast_fading_condition);


	for (int n = 0; n < N; n++)
	{
		delete[] GCS_phi_AOA[n];
		delete[] GCS_phi_AOD[n];
		delete[] GCS_theta_ZOA[n];
		delete[] GCS_theta_ZOD[n];

	}
	
	delete[] GCS_phi_AOA;
	delete[] GCS_phi_AOD;
	delete[] GCS_theta_ZOA;
	delete[] GCS_theta_ZOD;

}

void Generate_delays(double K,double DS, double *tau,double *delay, int num_cluster, int fast_fading_condition)
{
	double X;
	double C_tau;
	double r_tau = delay_scaling_param[fast_fading_condition];
	double *delay_prime;
	delay_prime = new double[num_cluster];

	for (int n = 0; n < num_cluster; n++)
	{
		if (calibration == true)
		{
			X = log(0.1 + n * 0.05);
		}
		else
		{
			X = log(Rand.u());
		}
		
		delay_prime[n] = - r_tau * DS *X;  
	}
	
	sort(delay_prime, delay_prime + num_cluster);
		
	for (int n = 0; n < num_cluster; n++)
	{
		tau[n] = delay_prime[n] - delay_prime[0];
	}


	if (fast_fading_condition == LOS)
	{
		C_tau = 0.7705 - 0.0433*K + 0.0002*pow(K, 2) + 0.000017*pow(K, 3);

			
		for (int n = 0; n < num_cluster; n++)
		{
			delay[n] = tau[n] / C_tau;
		}
	}
	else
	{
		for (int n = 0; n < num_cluster; n++)
		{
			delay[n] = tau[n] ;
			
		}
		
	}

	delete[] delay_prime;
}

/*===================================================================
FUNCTION: Generate_cluster_powers()

AUTHOR: GJ

DESCRIPTION: 

NOTES:

OUTPUT : num_real_cluster, delay[real_cluster], P[real_cluster]

===================================================================*/
int Generate_cluster_powers(int ue_idx,int bs_idx, double K_R, double P_1LOS, double DS, double *tau, double *delay, int num_cluster, int fast_fading_condition)
{
	double *P_prime;
	double Z;
	double r_tau = delay_scaling_param[fast_fading_condition];
	double zeta = per_cluster_showing_std[fast_fading_condition];
	double sum_P_prime = 0;
	double max_P;
	double Power[20];
	int num_real_cluster=0;
	P_prime = new double[num_cluster];
	double power_treshold = pow(10, -2.5); // -25dB

	//cout << "r_tau : " << r_tau << endl;
	//cout << "num_cluster : " << num_cluster << endl;
	//cout << "zeta : "<< zeta << endl;
	//cout << "fast_fading_condition : " << fast_fading_condition << endl;
	

	for (int n = 0; n < num_cluster; n++)
	{
		if (calibration == true)
		{
			Z = zeta;
		}
		else
		{
			Z = zeta * Rand.n();

		}
				
		P_prime[n] = exp(-tau[n] * (r_tau - 1) / (r_tau*DS)) *pow(10, -Z / 10);

		sum_P_prime += P_prime[n];
	}
	

	if (fast_fading_condition == LOS)
	{
		for (int n = 0; n < num_cluster; n++)
		{
			Power[n] = 1/(K_R+1) * P_prime[n]/sum_P_prime;
			if (n == 0)
			{
				Power[n]+= P_1LOS;
			}
		}
	}
	else
	{
		for (int n = 0; n < num_cluster; n++)
		{
			Power[n] = P_prime[n] / sum_P_prime;
		}
	}

	//for (int n = 0; n < num_cluster; n++)
	//{
	//	//cout << Power[n] << endl;
	//}

	max_P = max(Power, num_cluster);
	
	
	int m = 0;


	for (int n = 0 ; n < num_cluster; n++)
	{
		if (Power[n]/max_P>power_treshold) // 파워 수정
		{
			channel[bs_idx][ue_idx].cluster_power[m] = Power[n];
			channel[bs_idx][ue_idx].delay[m] = delay[n];
		
			num_real_cluster++;
			m++;
		}
	}
	
	delete[] P_prime;
	return num_real_cluster;
}

void Generate_arrival_departure_angle(int ue_idx,int bs_idx, double K, double ASA, double ASD, double ZSA, double ZSD,  double mu_offset_ZOD, int N, int M, int fast_fading_condition, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD, double phi_LOS_AOA, double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD)
{
	double C_phi;
	double C_phi_NLOS;
	double C_theta;
	double C_theta_NLOS;

	double X_AOA[20];
	double X_AOD[20];

	double Y_AOA[20];
	double Y_AOD[20];

	double X_ZOA[20];
	double X_ZOD[20];
	double Y_ZOA[20];
	double Y_ZOD[20];
	
	double alpha;
	double c_ASA = C_ASA[fast_fading_condition];
	double c_ASD = C_ASD[fast_fading_condition];
	double c_ZSA = C_ZSA[fast_fading_condition];
	double phi_n_AOA[20];
	double phi_n_AOD[20];
	double phi_prime_AOA[20];
	double phi_prime_AOD[20];
	double theta_n_ZOA[20];
	double theta_n_ZOD[20];
	double theta_prime_ZOA[20];
	double theta_prime_ZOD[20];

	double P_max;
	P_max = max(channel[bs_idx][ue_idx].cluster_power, N);

	
	if (AOD_AOA_distribution == Wrapped_Gaussian)
	{
		switch (num_cluster[fast_fading_condition])
		{
		case 4:
			C_phi_NLOS = 0.779;
			break;
		case 5:
			C_phi_NLOS = 0.860;
			break;
		case 8:
			C_phi_NLOS = 1.018;
			break;
		case 10:
			C_phi_NLOS = 1.090;
			break;
		case 11:
			C_phi_NLOS = 1.123;
			break;
		case 12:
			C_phi_NLOS = 1.146;
			break;
		case 14:
			C_phi_NLOS = 1.190;
			break;
		case 15:
			C_phi_NLOS = 1.211;
			break;
		case 16:
			C_phi_NLOS = 1.226;
			break;
		case 19:
			C_phi_NLOS = 1.273;
			break;
		case 20:
			C_phi_NLOS = 1.289;
			break;
		default:
			cout << "check num_cluster" << endl;
			break;
		}
		
		if (fast_fading_condition == LOS)
		{
			C_phi = C_phi_NLOS * (1.1035 - 0.028*K - 0.002*pow(K, 2) + 0.0001*pow(K, 3));
		}
		else
		{
			C_phi = C_phi_NLOS;
		}
	
		for (int n = 0; n < N; n++)
		{
			phi_prime_AOA[n] = 2. * (ASA / 1.4)*pow(-log(channel[bs_idx][ue_idx].cluster_power[n] / P_max), 0.5) / C_phi;
			phi_prime_AOD[n] = 2. * (ASD / 1.4)*pow(-log(channel[bs_idx][ue_idx].cluster_power[n] / P_max), 0.5) / C_phi;			
		}
	
	}
	else if(AOD_AOA_distribution == Laplacian)
	{
		switch (num_cluster[fast_fading_condition])
		{
		case 15:
			C_phi_NLOS = 1.434;
			break;
		case 19:
			C_phi_NLOS = 1.501;
			break;
		default:
			cout << "check num_cluster" << endl;
			cout <<"num_cluster[fast_fading_condition] = " << num_cluster[fast_fading_condition] << endl;
			getchar();
			break;
		}

		if (fast_fading_condition == LOS)
		{
			C_phi = C_phi_NLOS * (0.9275 + 0.0439*K - 0.0071*pow(K, 2) + 0.0002*pow(K, 3));
		}
		else
		{
			C_phi = C_phi_NLOS;
		}

		for (int n = 0; n < N; n++)
		{
			phi_prime_AOA[n] = -ASA * log(channel[bs_idx][ue_idx].cluster_power[n] / P_max) / C_phi;
			phi_prime_AOD[n] = -ASD * log(channel[bs_idx][ue_idx].cluster_power[n] / P_max) / C_phi;

		}

	}
	else
	{
		cout << "check void Generate_arrival_departure_angle()" << endl;
		getchar();
	}


	for (int n = 0; n < N; n++)
	{
		if (Rand.u() > 0.5)	X_AOA[n] = 1;
		else 	X_AOA[n] = -1;

		if (Rand.u() > 0.5)	X_AOD[n] = 1;
		else 	X_AOD[n] = -1;

		Y_AOA[n] = ASA / 7 * Rand.n();
		Y_AOD[n] = ASD / 7 * Rand.n();
	}

	for (int n = 0 ; n < N; n++)
	{
		
		if (fast_fading_condition == LOS)
		{
			if (n == 0)
			{
				phi_n_AOA[0] = phi_LOS_AOA;
				phi_n_AOD[0] = phi_LOS_AOD;
			}
			else
			{
				phi_n_AOA[n] = (X_AOA[n] * phi_prime_AOA[n] + Y_AOA[n]) - (X_AOA[0]*phi_prime_AOA[0] + Y_AOA[0] - phi_LOS_AOA);
				phi_n_AOD[n] = (X_AOD[n] * phi_prime_AOD[n] + Y_AOD[n]) - (X_AOD[0]*phi_prime_AOD[0] + Y_AOD[0] - phi_LOS_AOD);
			}
			
		}
		else
		{
			phi_n_AOA[n] = X_AOA[n] * phi_prime_AOA[n] +Y_AOA[n] + phi_LOS_AOA;
			phi_n_AOD[n] = X_AOD[n] * phi_prime_AOD[n] + Y_AOD[n] + phi_LOS_AOD;

			//cout << "phi_prime_AOD[n] : "<< phi_prime_AOD[n] << endl;
			
		}

		

		for (int m = 0; m < M; m++)
		{
			if (m < 2)
			{
				alpha = 0.0447;
			}
			else if (m<4)
			{
				alpha = 0.1413;
			}
			else if (m<6)
			{
				alpha = 0.2492;
			}
			else if (m<8)
			{
				alpha = 0.3715;
			}
			else if (m<10)
			{
				alpha = 0.5129;
			}
			else if (m<12)
			{
				alpha = 0.6797;
			}
			else if (m<14)
			{
				alpha = 0.8844;
			}
			else if (m<16)
			{
				alpha = 1.1481;
			}
			else if (m<18)
			{
				alpha = 1.5195;
			}
			else if (m<20)
			{
				alpha = 2.1551;
			}
			else
			{
				cout << "check num_rays_per_cluster" << endl;
				getchar();
			}

			if ( (m%2) == 1 )
			{
				alpha = -alpha;
			}

			phi_AOA[n][m] = phi_n_AOA[n] + c_ASA * alpha;
			phi_AOD[n][m] = phi_n_AOD[n] + c_ASD * alpha;
			
		}

	}



	////////////////////////////////// generate ZOA, ZOD//////////////////////////////////////////
	if (pathloss_model == InH_A || pathloss_model == UMa_A || pathloss_model == UMi_A || pathloss_model == RMa_A)
	{
		switch (num_cluster[fast_fading_condition])
		{
		case 10:
			C_theta_NLOS = 0.9854;
			break;
		case 11:
			C_theta_NLOS = 1.013;
			break;
		case 12:
			C_theta_NLOS = 1.04;
			break;
		case 15:
			C_theta_NLOS = 1.1088;
			break;
		case 19:
			C_theta_NLOS = 1.1764;
			break;
		case 20:
			C_theta_NLOS = 1.1918;
			break;
		default:
			cout << "check num_cluster" << endl;
			break;
		}

		if (fast_fading_condition == LOS)
		{
			C_theta = C_theta_NLOS * (1.35 + 0.0202*K - 0.0077*pow(K, 2) + 0.0002*pow(K, 3));
		}
		else
		{
			C_theta = C_theta_NLOS;
		}

	}
	else if(pathloss_model == InH_B || pathloss_model == UMa_B || pathloss_model == UMi_B || pathloss_model == RMa_B)
	{
		switch (num_cluster[fast_fading_condition])
		{
		case 8:
			C_theta_NLOS = 0.889;
			break;
		case 10:
			C_theta_NLOS = 0.957;
			break;
		case 11:
			C_theta_NLOS = 1.031;
			break;
		case 12:
			C_theta_NLOS = 1.104;
			break;
		case 15:
			C_theta_NLOS = 1.1088;
			break;
		case 19:
			C_theta_NLOS = 1.184;
			break;
		case 20:
			C_theta_NLOS = 1.178;
			break;
		default:
			cout << "check num_cluster" << endl;
			break;
		}
		if (fast_fading_condition == LOS)
		{
			C_theta = C_theta_NLOS * (1.3086 + 0.0339*K - 0.0077 *pow(K, 2) + 0.0002*pow(K, 3));
		}
		else
		{
			C_theta = C_theta_NLOS;
		}
	}
	else
	{
		cout << "check pathloss_model" << endl;
		getchar();
	}
	

	for (int n = 0; n < N; n++)
	{
		theta_prime_ZOA[n] = -ZSA * log(channel[bs_idx][ue_idx].cluster_power[n] / P_max) / C_theta;
		theta_prime_ZOD[n] = -ZSD* log(channel[bs_idx][ue_idx].cluster_power[n] / P_max) / C_theta; //XXX
		
	}
	
	

	for (int n = 0; n < N; n++)
	{
		if (Rand.u() > 0.5)	X_ZOA[n] = 1;
		else 	X_ZOA[n] = -1.;

		if (Rand.u() > 0.5)	X_ZOD[n] = 1;
		else 	X_ZOD[n] = -1.;

		Y_ZOA[n] = ZSA / 7.0 * Rand.n();
		Y_ZOD[n] = ZSD / 7.0 * Rand.n();
	}


	for (int n = 0; n < N; n++)
	{
		
		if (fast_fading_condition == LOS)
		{
			if (n == 0)
			{
				theta_n_ZOA[0] = theta_LOS_ZOA;
				theta_n_ZOD[0] = theta_LOS_ZOD;
			}
			else
			{
				theta_n_ZOA[n] = (X_ZOA[n] * theta_prime_ZOA[n] + Y_ZOA[n]) - (X_ZOA[0]*theta_prime_ZOA[0] + Y_ZOA[0] - theta_LOS_ZOA);
				theta_n_ZOD[n] = (X_ZOD[n] * theta_prime_ZOD[n] + Y_ZOD[n]) - (X_ZOD[0]*theta_prime_ZOD[0] + Y_ZOD[0] - theta_LOS_ZOD);
			}			
		}
		else
		{
			if (fast_fading_condition == OUT2IN)
			{
				theta_LOS_ZOA = 90;
			}

			theta_n_ZOA[n] = X_ZOA[n] * theta_prime_ZOA[n] +Y_ZOA[n] + theta_LOS_ZOA; 	

			theta_n_ZOD[n] = X_ZOD[n] * theta_prime_ZOD[n] + Y_ZOD[n] + theta_LOS_ZOD + mu_offset_ZOD;
			

		}

		
		
		for (int m = 0; m < M; m++)
		{
			if (m < 2)
			{
				alpha = 0.0447;
			}
			else if (m<4)
			{
				alpha = 0.1413;
			}
			else if (m<6)
			{
				alpha = 0.2492;
			}
			else if (m<8)
			{
				alpha = 0.3715;
			}
			else if (m<10)
			{
				alpha = 0.5129;
			}
			else if (m<12)
			{
				alpha = 0.6797;
			}
			else if (m<14)
			{
				alpha = 0.8844;
			}
			else if (m<16)
			{
				alpha = 1.1481;
			}
			else if (m<18)
			{
				alpha = 1.5195;
			}
			else if (m<20)
			{
				alpha = 2.1551;
			}
			else
			{
				cout << "check num_rays_per_cluster" << endl;
				getchar();
			}

			if ((m % 2) == 1)
			{
				alpha = -alpha;
			}

			theta_ZOA[n][m] = theta_n_ZOA[n] + c_ZSA * alpha;
			
			if (pathloss_model == InH_A || pathloss_model == InH_B)
			{
				theta_ZOD[n][m] = theta_n_ZOD[n] + 3. / 8. *pow(10,log_ZSD_mean[fast_fading_condition])*alpha; 
			}
			else
			{
				theta_ZOD[n][m] =theta_n_ZOD[n] + 3. / 8.*pow(10, channel[bs_idx][ue_idx].log_ZSD_mean[fast_fading_condition])*alpha; 
			}
			
		}

	}


	for (int n = 0; n < N; n++)
	{
		for (int m = 0; m < M; m++)
		{
			phi_AOD[n][m] = Transform_angle_minus_180_to_plus_180(phi_AOD[n][m]);
			theta_ZOD[n][m] = Transform_angle_0_to_plus_180(theta_ZOD[n][m]);
			
			phi_AOA[n][m] = Transform_angle_minus_180_to_plus_180(phi_AOA[n][m]);
			theta_ZOA[n][m] = Transform_angle_0_to_plus_180(theta_ZOA[n][m]);
		}
	}

	
	
}

void Couple_rays(int ue_idx, int bs_idx,  int N, int M, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD)
{
	int AOD_coupling_ray[20];
	int AOA_coupling_ray[20];
	int ZOD_coupling_ray[20];
	int ZOA_coupling_ray[20];
	
	double temp_phi_AOA[20];
	double temp_phi_AOD[20];
	double temp_theta_ZOA[20];
	double temp_theta_ZOD[20];

	Find_strong_cluster(ue_idx, bs_idx, N);
	
	for (int cluster_idx = 0; cluster_idx < N; cluster_idx++)
	{
		if ((cluster_idx == channel[bs_idx][ue_idx].strong_cluster[0]) || (cluster_idx == channel[bs_idx][ue_idx].strong_cluster[1]))
		{
			Shuffle_strong_ray(AOD_coupling_ray);
			Shuffle_strong_ray(AOA_coupling_ray);

			Shuffle_strong_ray(ZOD_coupling_ray);
			Shuffle_strong_ray(ZOA_coupling_ray);
			
			/*
			for (int ray_idx = 0; ray_idx<M; ray_idx++)
			{
				cout << AOD_coupling_ray[ray_idx] << " ";
				if ((ray_idx == 9)|| (ray_idx == 15))
				{
					cout << "   " << endl;
				}
			}
			cout << endl;
			for (int ray_idx = 0; ray_idx<M; ray_idx++)
			{
				cout << AOA_coupling_ray[ray_idx] << " ";
				if ((ray_idx == 9) || (ray_idx == 15))
				{
					cout << "   " << endl;
				}
			}
			cout << endl;
			for (int ray_idx = 0; ray_idx<M; ray_idx++)
			{
				cout << ZOA_coupling_ray[ray_idx] << " ";
				if ((ray_idx == 9) || (ray_idx == 15))
				{
					cout << "   " << endl;
				}
			}
			cout << endl;
			for (int ray_idx = 0; ray_idx<M; ray_idx++)
			{
				cout << ZOD_coupling_ray[ray_idx] << " ";
				if ((ray_idx == 9) || (ray_idx == 15))
				{
					cout << "   " << endl;
				}
			}
			cout << endl;
			getchar();
			*/
		}
		else
		{
			Shuffle_weak_ray(AOD_coupling_ray);
			Shuffle_weak_ray(AOA_coupling_ray);
			Shuffle_weak_ray(ZOD_coupling_ray);
			Shuffle_weak_ray(ZOA_coupling_ray);
				

			/*
			cout << endl;
			cout << "AOA_coupling_ray : ";
			for (int ray_idx = 0; ray_idx < M; ray_idx++)
			{
				cout << AOD_coupling_ray[ray_idx] << " ";
			}
			cout << endl;

			cout << "AOA_coupling_ray : ";
			for (int ray_idx = 0; ray_idx < M; ray_idx++)
			{
				cout << AOA_coupling_ray[ray_idx] << " ";
			}

			cout << endl;

			cout << "ZOD_coupling_ray : ";
			for (int ray_idx = 0; ray_idx < M; ray_idx++)
			{
				cout << ZOD_coupling_ray[ray_idx] << " ";
			}cout << endl;

			cout << "ZOA_coupling_ray : ";
			for (int ray_idx = 0; ray_idx < M; ray_idx++)
			{
				cout << ZOA_coupling_ray[ray_idx] << " ";
			}

			getchar();
			*/
		}

		for (int ray_idx = 0; ray_idx < M; ray_idx++)
		{
			temp_phi_AOA[ray_idx] = phi_AOA[cluster_idx][ray_idx];
			temp_phi_AOD[ray_idx] = phi_AOD[cluster_idx][ray_idx];
			temp_theta_ZOA[ray_idx] = theta_ZOA[cluster_idx][ray_idx];
			temp_theta_ZOD[ray_idx] = theta_ZOD[cluster_idx][ray_idx];
		}

		for (int ray_idx = 0; ray_idx < M; ray_idx++)
		{
			
			phi_AOA[cluster_idx][ray_idx] = temp_phi_AOA[AOA_coupling_ray[ray_idx]];
			phi_AOD[cluster_idx][ray_idx] = temp_phi_AOD[AOD_coupling_ray[ray_idx]];
			
			theta_ZOA[cluster_idx][ray_idx] = temp_theta_ZOA[ZOA_coupling_ray[ray_idx]];
			theta_ZOD[cluster_idx][ray_idx] = temp_theta_ZOD[ZOD_coupling_ray[ray_idx]];
		}
	/*	cout << endl;
		cout << "temp_phi_ZOD : " << endl;

		for (int ray_idx = 0; ray_idx < M; ray_idx++)
		{
			cout << ZOD_coupling_ray[ray_idx] << " ";
		}

		cout << endl;
		for (int ray_idx = 0; ray_idx < M; ray_idx++)
		{
			cout << ray_idx <<"  "<< temp_theta_ZOD[ray_idx] << "  " << theta_ZOD[cluster_idx][ray_idx] << endl;
		} 

		
		getchar();*/

	}

}



void Generate_XPR(int ue_idx, int bs_idx, double mu_XPR, double sigma_XPR, int N, int M)
{
	for (int n=0; n<N; n++)
	{
		for (int m = 0; m<M; m++)
		{
			channel[bs_idx][ue_idx].kappa[n][m] = pow(10., (sigma_XPR*Rand.n() + mu_XPR) / 10.);
		}
	}

}



/*===================================================================
FUNCTION: Generate_channel_coef_elements()

AUTHOR: GJ

DESCRIPTION:

M.2412 page 58~60

Generate channel coefficents between antenna port except doppler shift. 



NOTES: Output is a Channel_Coef.

===================================================================*/

void Generate_channel_coef_elements(int ue_idx,int bs_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double  GCS_theta_LOS_ZOD, int N, int M, int fast_fading_condition)
{

	Generate_initial_random_phase(ue_idx, bs_idx, N, M);

	if (fast_fading_condition == LOS)
	{
		Generate_LOS_field(ue_idx, bs_idx,GCS_phi_LOS_AOA, GCS_phi_LOS_AOD, GCS_theta_LOS_ZOA, GCS_theta_LOS_ZOD); 	// Output : GCS_antenna_element_LOS_departure_field[2][2], GCS_antenna_element_LOS_arrival_field[3][2]

		Generate_LOS_wave_direction_vector(ue_idx, bs_idx, GCS_phi_LOS_AOA, GCS_phi_LOS_AOD, GCS_theta_LOS_ZOA, GCS_theta_LOS_ZOD);

		Generate_LOS_UE_velocity(ue_idx, bs_idx, GCS_phi_LOS_AOA, GCS_phi_LOS_AOD, GCS_theta_LOS_ZOA, GCS_theta_LOS_ZOD);
	}

	Generate_NLOS_field(ue_idx, bs_idx, GCS_phi_AOA, GCS_phi_AOD, GCS_theta_ZOA, GCS_theta_ZOD,  N,  M); 	// Output : GCS_antenna_element_NLOS_departure_field[20][20][2][2][2][2], GCS_antenna_element_NLOS_arrival_field[20][20][3][2]
	
	Generate_NLOS_wave_direction_vector(ue_idx, bs_idx, GCS_phi_AOA, GCS_phi_AOD, GCS_theta_ZOA, GCS_theta_ZOD, N, M);

	Generate_NLOS_UE_velocity(ue_idx, bs_idx, GCS_phi_AOA, GCS_phi_AOD, GCS_theta_ZOA, GCS_theta_ZOD, N, M);

}


/*===================================================================
FUNCTION: void Generate_LOS_field

AUTHOR: GJ

DESCRIPTION: Generate_LOS_field

INPUT :  

- wave GCS_angle : GCS_phi_LOS_AOA, GCS_phi_LOS_AOD, GCS_theta_LOS_ZOA, GCS_theta_LOS_ZOD
- rotation angle : 
ue[ue_idx].alpha,beta,gamma
sector_alpha,beta,gamma

ue_alpha -> random_alpah, random_alpah+180  2pannel antenna
sectoer_alpha -> 30 150 270 

OUTPUT : 

- GCS_field

GCS_LOS_departure_field[2][2].phi
GCS_LOS_departure_field[2][2].theta

GCS_LOS_arrival_field[3][2].phi
GCS_LOS_arrival_field[3][2].theta

first array : 
# of pannel in ue(bearing angle = alpha, alpha+PI)
# of secter (bearing anlge = 30, 150, 270)

second array ;
# of polarization 2

omnidirectional antenna case : save GCS_LOS_departure_field[0][0],GCS_LOS_departure_field[0][1]

NOTES: Currently, nothing is done when a sector is instantiated

===================================================================*/

void Generate_initial_random_phase(int ue_idx, int bs_idx, int  N, int M)
{
	complex <double> j(0.0, 1.0);

	double PHI_LOS = -2.*PI*channel[bs_idx][ue_idx]._3D_distance / lamda_0; // m.2412
	channel[bs_idx][ue_idx].exp_j_PHI_LOS = exp(j*PHI_LOS);

	for (int n=0;n<N;n++)
	{
		for (int m = 0; m<M; m++)
		{
			channel[bs_idx][ue_idx].exp_j_PHI[n][m].theta_theta = exp(j*(2.0*PI*Rand.u() - PI));
			channel[bs_idx][ue_idx].exp_j_PHI[n][m].theta_phi = exp(j*(2.0*PI*Rand.u() - PI));
			channel[bs_idx][ue_idx].exp_j_PHI[n][m].phi_theta = exp(j*(2.0*PI*Rand.u() - PI));
			channel[bs_idx][ue_idx].exp_j_PHI[n][m].phi_phi = exp(j*(2.0*PI*Rand.u() - PI));
		}
	}
	
}

void Generate_LOS_field(int ue_idx, int bs_idx, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double GCS_theta_LOS_ZOD)
{
	SPHERICAL_ANGLE GCS_LOS_departure;
	SPHERICAL_ANGLE GCS_LOS_arrival;
	
	SPHERICAL_ANGLE LCS_LOS_depature[2]; // 2 bearing angle alpha , alpah + 180
	SPHERICAL_ANGLE LCS_LOS_arrival[3];  // 3 sector

	ANTENNA_FIELD LCS_departure_field[2]; // for omni directional antenna
	ANTENNA_FIELD LCS_pannel_antenna_departure_field[2][2];// first term : bearing angle 2 pannel  *** second term : polarization 
	ANTENNA_FIELD LCS_sector_arrival_field[3][2]; // first term : bearing angle 3 sector ,  *** second term : polarization 
	

	double ue_antenna_gain_dB;
	double ue_pannel_antenna_gain_dB[2]; // opposite pannel antenna
	double sector_antenna_gain_dB[3];
		
	double ue_alpha, ue_beta, ue_gamma;
	double sector_alpha[3], sector_beta, sector_gamma;
	
	double ue_zeta[2] = {0, PI / 2 };
	
	double sector_zeta[2] = { -PI / 4, PI / 4 };
	

	GCS_LOS_departure.theta = GCS_theta_LOS_ZOD;
	GCS_LOS_departure.phi = GCS_phi_LOS_AOD;

	GCS_LOS_arrival.theta = GCS_theta_LOS_ZOA;
	GCS_LOS_arrival.phi = GCS_phi_LOS_AOA;



	ue_alpha =  ue[ue_idx].alpha;
	ue_beta = ue[ue_idx].beta;
	ue_gamma = ue[ue_idx].gamma;
	
	if (num_TRxP_per_site == 3)
	{
		sector_alpha[0] = 30. * PI / 180.;
		sector_alpha[1] = (30. + 120.) * PI / 180.;
		sector_alpha[2] = (30. + 240.) * PI / 180.;
	}
	else
	{
		sector_alpha[0] = 0.;
	}
	
	sector_beta  = (bs_antenna.mechanic_tilt - 90.) * PI / 180.;
	sector_gamma = 0.;


	
	//********* Departure Antenna_field calculation*****************//
	if (ue_antenna.element_gain_dBi == 0)
	{
		ue_antenna_gain_dB = 0;

		for (int pol_idx = 0; pol_idx < 2; pol_idx++)
		{
			LCS_departure_field[pol_idx] = Antenna_field_in_LCS(ue_antenna_gain_dB, ue_zeta[pol_idx]);

			channel[bs_idx][ue_idx].GCS_LOS_departure_field[0][pol_idx] = LCS_Antenna_field_to_GCS_antenna_pattern(0, 0, 0, GCS_LOS_departure, LCS_departure_field[pol_idx]); // first term is 0 -> only 1 pannel

			/*cout << "channel[bs_idx][ue_idx].GCS_LOS_departure_field[0][pol_idx].phi :" << channel[bs_idx][ue_idx].GCS_LOS_departure_field[0][pol_idx].phi << endl;
			cout << "channel[bs_idx][ue_idx].GCS_LOS_departure_field[0][pol_idx]].theta  : " << channel[bs_idx][ue_idx].GCS_LOS_departure_field[0][pol_idx].theta << endl;
			getchar();*/
		}
	}
	else
	{
		for (int pannel_idx = 0; pannel_idx < 2; pannel_idx++)
		{
			// GCS angle-> LCS angle : 2 bearing angle (2 pannel)

			LCS_LOS_depature[pannel_idx] = GCS_Angle_to_LCS_Angle( ue_alpha + pannel_idx*PI, ue_beta, ue_gamma, GCS_LOS_departure);
		
			// Antenna gain in LCS :  2 bearing angle (2 pannel)

			ue_pannel_antenna_gain_dB[pannel_idx] = UE_Antenna_gain_in_dB(LCS_LOS_depature[pannel_idx]);	

			for (int pol_idx = 0; pol_idx < 2; pol_idx++)
			{
				// LCS Antenna field calculation 2 pannel , 2 polarization
				LCS_pannel_antenna_departure_field[pannel_idx][pol_idx] = Antenna_field_in_LCS(ue_pannel_antenna_gain_dB[pannel_idx], ue_zeta[pol_idx]);
	
				// Transform LCS field to GCS antenna field
				channel[bs_idx][ue_idx].GCS_LOS_departure_field[pannel_idx][pol_idx] = LCS_Antenna_field_to_GCS_antenna_pattern(ue_alpha + pannel_idx * PI, ue_beta, ue_gamma, GCS_LOS_departure, LCS_pannel_antenna_departure_field[pannel_idx][pol_idx]);
			}
		}
	}
	
	
	
	//********* Arrival Antenna_field calculation*****************//
	
	//cout << endl;
	//cout <<"-------------------------------"<< endl;
	//cout << "ue_idx : " << ue_idx << endl;
	//cout << "bs_idx : " << bs_idx << endl;
	for (int sec_idx_in_bs = 0; sec_idx_in_bs < num_TRxP_per_site; sec_idx_in_bs++)
	{
		// GCS angle-> LCS angle : 3 bearing angle
		/*cout<<endl;
		cout <<"_2D_distance : " <<channel[bs_idx][ue_idx]._2D_distance << endl;
		cout <<"GCS_LOS_arrival.phi : "<< GCS_LOS_arrival.phi << endl;
		cout <<"GCS_LOS_arrival.thata : "<< GCS_LOS_arrival.theta << endl;
*/

		LCS_LOS_arrival[sec_idx_in_bs] = GCS_Angle_to_LCS_Angle(sector_alpha[sec_idx_in_bs], sector_beta, sector_gamma, GCS_LOS_arrival);
		
		sector_antenna_gain_dB[sec_idx_in_bs] = Sector_Antenna_gain_in_dB(LCS_LOS_arrival[sec_idx_in_bs]);

		for (int pol_idx = 0; pol_idx < 2; pol_idx++)
		{
			// Antenna field in LCS : 3 bearing angle, 2 polarization, 2 component in LCS the direction spherical basis vector theta, pi
			
			LCS_sector_arrival_field[sec_idx_in_bs][pol_idx] = Antenna_field_in_LCS(sector_antenna_gain_dB[sec_idx_in_bs], sector_zeta[pol_idx]);

			// LCS antenna field - > GCS antenna field : 3 bearing angle, 2 polarization, GCS 2 component in the direction spherical basis vector theta, pi
			// ==============Output ========================//
			
			channel[bs_idx][ue_idx].GCS_LOS_arrival_field[sec_idx_in_bs][pol_idx] = LCS_Antenna_field_to_GCS_antenna_pattern(sector_alpha[sec_idx_in_bs], sector_beta, sector_gamma, GCS_LOS_arrival, LCS_sector_arrival_field[sec_idx_in_bs][pol_idx]);
		}

	}
	//getchar();
}
void Generate_NLOS_field(int ue_idx, int bs_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD,int N, int M)
{
	SPHERICAL_ANGLE GCS_departure[20][20];
	SPHERICAL_ANGLE GCS_arrival[20][20];

	SPHERICAL_ANGLE LCS_depature[2][20][20]; // 2 bearing angle alpha , alpah + 180
	SPHERICAL_ANGLE LCS_arrival[3][20][20];  // 3 sector

	ANTENNA_FIELD LCS_departure_field[2][20][20];
	ANTENNA_FIELD LCS_pannel_antenna_departure_field[2][2][20][20];// first term : 2 pannel ,  second term : polarization 
	ANTENNA_FIELD LCS_sector_arrival_field[3][2][20][20]; //          first term : 3 sector ,  second term : polarization 


	double ue_antenna_gain_dB;
	double ue_pannel_antenna_gain_dB[2][20][20]; // opposite pannel antenna
	double sector_antenna_gain_dB[3][20][20];

	double ue_alpha, ue_beta, ue_gamma;
	double sector_alpha[3], sector_beta, sector_gamma;

	double ue_zeta[2] = { 0, PI / 2 };

	double sector_zeta[2] = { -PI / 4, PI / 4 };

	//cout << "GCS_Departure_angle_LOS.phi : "<<channel[bs_idx][ue_idx].GCS_Departure_angle_LOS.phi << endl;
	//cout << N << endl;
	for (int n=0; n<N; n++)
	{
		for (int m = 0; m < M; m++)
		{
	
			GCS_departure[n][m].phi = GCS_phi_AOD[n][m];
			GCS_departure[n][m].theta = GCS_theta_ZOD[n][m];

			GCS_arrival[n][m].phi = GCS_phi_AOA[n][m];
			GCS_arrival[n][m].theta = GCS_theta_ZOA[n][m];
	
			//cout << GCS_arrival[n][m].phi << endl;
		}	
		//getchar();
	}
	
	
	ue_alpha = ue[ue_idx].alpha;
	ue_beta = ue[ue_idx].beta;
	ue_gamma = ue[ue_idx].gamma;

	if (num_TRxP_per_site == 3)
	{
		sector_alpha[0] = (30) * PI / 180;
		sector_alpha[1] = (30 + 120) * PI / 180;
		sector_alpha[2] = (30 + 240) * PI / 180;
	}
	else
	{
		sector_alpha[0] = 0;
	}
	sector_beta = (bs_antenna.mechanic_tilt - 90) * PI / 180;
	sector_gamma = 0;


	//********* Departure Antenna_field calculation*****************//
	
	if (ue_antenna.element_gain_dBi == 0)
	{
		ue_antenna_gain_dB = 0;
		for (int pol_idx = 0; pol_idx < 2; pol_idx++)
		{
			for (int n = 0; n < N; n++)
			{
				for (int m = 0; m < M; m++)
				{
					LCS_departure_field[pol_idx][n][m] = Antenna_field_in_LCS(ue_antenna_gain_dB, ue_zeta[pol_idx]);
									
					channel[bs_idx][ue_idx].GCS_NLOS_departure_field[0][pol_idx][n][m] = LCS_Antenna_field_to_GCS_antenna_pattern(0, 0, 0, GCS_departure[n][m], LCS_departure_field[pol_idx][n][m]);	
				}
			}
		}
	}
	else
	{
		for (int pannel_idx = 0; pannel_idx < 2; pannel_idx++)
		{
			// GCS angle-> LCS angle : 2 bearing angle (2 pannel)

			for (int n = 0; n < N; n++)
			{
				for (int m = 0; m < M; m++)
				{
					
					LCS_depature[pannel_idx][n][m] = GCS_Angle_to_LCS_Angle(ue_alpha + double(pannel_idx) * PI, ue_beta, ue_gamma, GCS_departure[n][m]);

				
					// Antenna gain in LCS :  2 bearing angle (2 pannel)

					ue_pannel_antenna_gain_dB[pannel_idx][n][m] = UE_Antenna_gain_in_dB(LCS_depature[pannel_idx][n][m]);

					for (int pol_idx = 0; pol_idx < 2; pol_idx++)
					{
						// LCS Antenna field calculation 2 pannel , 2 polarization
						LCS_pannel_antenna_departure_field[pannel_idx][pol_idx][n][m] = Antenna_field_in_LCS(ue_pannel_antenna_gain_dB[pannel_idx][n][m], ue_zeta[pol_idx]);

						// Transform LCS field to GCS antenna field
						channel[bs_idx][ue_idx].GCS_NLOS_departure_field[pannel_idx][pol_idx][n][m] = LCS_Antenna_field_to_GCS_antenna_pattern(ue_alpha + pannel_idx * PI, ue_beta, ue_gamma, GCS_departure[n][m], LCS_pannel_antenna_departure_field[pannel_idx][pol_idx][n][m]);
						
					}
				}
			}
		
		}
	}

	//********* Arrival Antenna_field calculation*****************//

	// 3 sector


	for (int sec_idx_in_bs = 0; sec_idx_in_bs < num_TRxP_per_site; sec_idx_in_bs++)
	{
		for (int n = 0; n < N; n++)
		{
			for (int m = 0; m < M; m++)
			{
				// GCS angle-> LCS angle : 3 bearing angle 
				LCS_arrival[sec_idx_in_bs][n][m] = GCS_Angle_to_LCS_Angle(sector_alpha[sec_idx_in_bs], sector_beta, sector_gamma, GCS_arrival[n][m]);

				//안테나 게인 고정

				//LCS_arrival[sec_idx_in_bs][n][m].theta = 90;
				//LCS_arrival[sec_idx_in_bs][n][m].phi = 0;
				//
				// Antenna gain in LCS : 3 bearing angle
				sector_antenna_gain_dB[sec_idx_in_bs][n][m] = Sector_Antenna_gain_in_dB(LCS_arrival[sec_idx_in_bs][n][m]);
				
				for (int pol_idx = 0; pol_idx < 2; pol_idx++)
				{
					// Antenna field in LCS : 3 bearing angle, 2 polarization, 2 component in LCS the direction spherical basis vector theta, pi

					LCS_sector_arrival_field[sec_idx_in_bs][pol_idx][n][m] = Antenna_field_in_LCS(sector_antenna_gain_dB[sec_idx_in_bs][n][m], sector_zeta[pol_idx]);

					// LCS antenna field - > GCS antenna field : 3 bearing angle, 2 polarization, GCS 2 component in the direction spherical basis vector theta, pi
					// ==============Output ========================//
					channel[bs_idx][ue_idx].GCS_NLOS_arrival_field[sec_idx_in_bs][pol_idx][n][m] = LCS_Antenna_field_to_GCS_antenna_pattern(sector_alpha[sec_idx_in_bs], sector_beta, sector_gamma, GCS_arrival[n][m], LCS_sector_arrival_field[sec_idx_in_bs][pol_idx][n][m]);

				}
			}
		}

	}


}

void Generate_LOS_wave_direction_vector(int ue_idx,int bs_idx, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double GCS_theta_LOS_ZOD)
{
	channel[bs_idx][ue_idx].r_tx_LOS = Transform_angle_to_spheical_vector(GCS_phi_LOS_AOD, GCS_theta_LOS_ZOD); // tx

	
	channel[bs_idx][ue_idx].r_rx_LOS = Transform_angle_to_spheical_vector(GCS_phi_LOS_AOA, GCS_theta_LOS_ZOA); // rx

	/*channel[bs_idx][ue_idx].GCS_Arrival_angle_LOS.phi = GCS_phi_LOS_AOA;
	channel[bs_idx][ue_idx].GCS_Arrival_angle_LOS.theta = GCS_theta_LOS_ZOA;*/
	
}
void Generate_NLOS_wave_direction_vector(int ue_idx, int bs_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD, int N, int M)
{
	for (int n = 0; n<N; n++)
	{
		for (int m = 0; m < M; m++)
		{
			channel[bs_idx][ue_idx].r_tx[n][m] = Transform_angle_to_spheical_vector(GCS_phi_AOD[n][m], GCS_theta_ZOD[n][m]);
			channel[bs_idx][ue_idx].r_rx[n][m] = Transform_angle_to_spheical_vector(GCS_phi_AOA[n][m], GCS_theta_ZOA[n][m]);
		}
	}
}

void Generate_NLOS_UE_velocity(int ue_idx, int bs_idx, double **GCS_phi_AOA, double **GCS_phi_AOD, double **GCS_theta_ZOA, double **GCS_theta_ZOD, int N, int M)
{

	for (int n = 0; n<N; n++)
	{
		for (int m = 0; m < M; m++)
		{
			channel[bs_idx][ue_idx].velocity[n][m] = dot(ue[ue_idx].velocity, channel[bs_idx][ue_idx].r_rx[n][m]);
		}
		
	}
}
void Generate_LOS_UE_velocity(int ue_idx, int bs_idx, double GCS_phi_LOS_AOA, double  GCS_phi_LOS_AOD, double  GCS_theta_LOS_ZOA, double GCS_theta_LOS_ZOD)
{
	channel[bs_idx][ue_idx].velocity_LOS = dot(ue[ue_idx].velocity, channel[bs_idx][ue_idx].r_rx_LOS);
}



double max(double *A, int size)
{
	double max_value=0;
	for (int n = 0; n< size; n++)
	{
		if (n == 0)
		{
			max_value = A[0];
		}
		else 
		{
			if (A[n] > max_value)
			{
				max_value = A[n];
			}
		}
	}
	return max_value;
}

double Transform_angle_minus_180_to_plus_180(double x)
{
	
	x = fmod(x, 360);

	if (x > 180)
	{
		x = x - 360;
	}
	else if (x < -180)
	{
		x = x + 360;
	}

	return x;
}

double Transform_angle_0_to_plus_180(double x)
{
	x = fmod(x, 360);
	if (x > 180)
	{
		x = 360 - x;
	}
	else if ((x < 0) && (x >= -180))
	{
		x = -x;
	}
	else if (x < -180)
	{
		x = x + 360;
	}
	return x;

}

LOCATION Transform_angle_to_spheical_vector(double phi ,double theta)
{
	LOCATION r;

	double Theta = theta / 180. * PI;
	double Phi = phi / 180. * PI;

	r.x = sin(Theta) * cos(Phi);
	r.y = sin(Theta) * sin(Phi);
	r.z = cos(Theta);


	return r;
}

void Find_strong_cluster(int ue_idx,int bs_idx, int N)
{
	int max_idx=0;
	int second_max_idx=0;
	double max_value;
	double second_max_value;
	
	
	for (int n = 1; n< N; n++)
	{
		
		if(n == 1)
		{
			if (channel[bs_idx][ue_idx].cluster_power[1] > channel[bs_idx][ue_idx].cluster_power[0])
			{
				max_value = channel[bs_idx][ue_idx].cluster_power[1];
				second_max_value= channel[bs_idx][ue_idx].cluster_power[0];
				max_idx = 1;
				second_max_idx = 0;
			}
			else
			{
				max_value = channel[bs_idx][ue_idx].cluster_power[0];
				second_max_value = channel[bs_idx][ue_idx].cluster_power[1];
				max_idx = 0;
				second_max_idx = 1;
			}

		}
		else
		{
			if (channel[bs_idx][ue_idx].cluster_power[n] > max_value)
			{
				max_value = channel[bs_idx][ue_idx].cluster_power[n];
				second_max_idx = max_idx;
				max_idx = n;			
			}
			else
			{
				if (channel[bs_idx][ue_idx].cluster_power[n] > second_max_value)
				{
					second_max_value= channel[bs_idx][ue_idx].cluster_power[n];
					second_max_idx = n;
				}

			}
		}
	}
		
	channel[bs_idx][ue_idx].strong_cluster[0] = max_idx;
	channel[bs_idx][ue_idx].strong_cluster[1] = second_max_idx;
	
	/*
	for (int n = 0; n < N; n++)
	{
		cout << " idx : " << n << "  P " <<  P[n] << endl;
	}

	cout<<"max : " << channel[bs_idx][ue_idx].strong_cluster[0] << endl;
	cout<<"second :" << channel[bs_idx][ue_idx].strong_cluster[1] << endl;

	getchar();
	*/

}


void Shuffle_weak_ray(int *rays)
{
	vector<int> a(20);
	for (int i = 0; i < 20; i++) a[i] = i;

	random_shuffle(a.begin(), a.end());
	for (int ray_idx = 0; ray_idx<20; ray_idx++)
	{
		rays[ray_idx] = a[ray_idx];
	}
}

void Shuffle_strong_ray(int *rays)
{
	vector<int> sub_cluster_1 = { 0,1,2,3,4,5,6,7,18,19 };
	vector<int> sub_cluster_2 = { 8,9,10,11,16,17 };
	vector<int> sub_cluster_3 = { 12,13,14,15 };

	random_shuffle(sub_cluster_1.begin(), sub_cluster_1.end());
	random_shuffle(sub_cluster_2.begin(), sub_cluster_2.end());
	random_shuffle(sub_cluster_3.begin(), sub_cluster_3.end());


	for (int ray_idx = 0; ray_idx<10; ray_idx++)
	{
		rays[ray_idx] = sub_cluster_1[ray_idx];
	}
	
	
	for (int ray_idx = 0; ray_idx<6; ray_idx++)
	{
		rays[ray_idx + 10] = sub_cluster_2[ray_idx];
	}
		
	for (int ray_idx = 0; ray_idx<4; ray_idx++)
	{
		rays[ray_idx + 16] = sub_cluster_3[ray_idx];
	}

}


SPHERICAL_ANGLE GCS_Angle_to_LCS_Angle(double alpha, double beta, double gamma, SPHERICAL_ANGLE GCS_angle) //GCS_angle : degree  ||  alpha, beta gamma : radian
{
	SPHERICAL_ANGLE LCS_angle;
	double theta = GCS_angle.theta * PI / 180.0;
	double phi = GCS_angle.phi * PI / 180.0;

	LCS_angle.theta = acosl(cos(beta) * cos(gamma) * cos(theta) + (sin(beta)*cos(gamma)*cos(phi - alpha) - sin(gamma)*sin(phi - alpha))*sin(theta));
	
	LCS_angle.phi = atan2l((cos(beta)*sin(gamma)*cos(theta) + (sin(beta)*sin(gamma)*cos(phi - alpha) + cos(gamma)*sin(phi - alpha))*sin(theta)), (cos(beta)*sin(theta)*cos(phi - alpha) - sin(beta)*cos(theta)));
	
	LCS_angle.theta = LCS_angle.theta * 180.0 / PI;

	LCS_angle.phi = LCS_angle.phi * 180.0 / PI;
	return LCS_angle;
}

double Sector_Antenna_gain_in_dB(SPHERICAL_ANGLE LCS_angle) //LCS_angle : degree
{
	double _3dB_angle;
	double SLA;
	double Am;
	double antenna_gain_Vertical;
	double antenna_gain_Horizontal;
	double antenna_gain;

	if (test_environment >= 3) // NOT INDOOR 
	{
		_3dB_angle =65.0;
		SLA = 30.0;
		Am = 30.0;
	}
	else// INDOOR
	{
		_3dB_angle = 90.0;
		SLA = 25.0;
		Am = 25.0;		
	}


	antenna_gain_Vertical = -min(12.0 * pow(((LCS_angle.theta - 90.0) / _3dB_angle), 2.0), SLA);
	antenna_gain_Horizontal = -min(12.0 * pow((LCS_angle.phi / _3dB_angle), 2.0), Am);
	antenna_gain = -(min(-(antenna_gain_Vertical + antenna_gain_Horizontal), Am));

	antenna_gain = antenna_gain + bs_antenna.element_gain_dBi;
	
	return antenna_gain;

}

double UE_Antenna_gain_in_dB(SPHERICAL_ANGLE LCS_angle) //LCS_angle : degree
{
	double _3dB_angle;
	double SLA;
	double Am;
	double antenna_gain_Vertical;
	double antenna_gain_Horizontal;
	double antenna_gain;

	_3dB_angle = 90.0;
	SLA = 25.0;
	Am = 25.0;

	antenna_gain_Vertical = -min(12.0 * pow(((LCS_angle.theta - 90.0) / _3dB_angle), 2.0), SLA);
	antenna_gain_Horizontal = -min(12.0 * pow((LCS_angle.phi / _3dB_angle), 2.0), Am);
	antenna_gain = -(min(-(antenna_gain_Vertical + antenna_gain_Horizontal), Am));

	antenna_gain = antenna_gain + ue_antenna.element_gain_dBi;

	return antenna_gain;

}

ANTENNA_FIELD Antenna_field_in_LCS(double LCS_antenna_gain_in_dB, double zeta)
{
	ANTENNA_FIELD field_pattern;
	double LCS_antenna_gain = dB2linear(LCS_antenna_gain_in_dB);


	field_pattern.theta = pow(LCS_antenna_gain,0.5)*cos(zeta);
	field_pattern.phi   = pow(LCS_antenna_gain,0.5)*sin(zeta);

	return field_pattern;
}

ANTENNA_FIELD LCS_Antenna_field_to_GCS_antenna_pattern(double alpha, double beta, double gamma, SPHERICAL_ANGLE GCS_angle, ANTENNA_FIELD LCS_field) //GCS_angle degree
{
	ANTENNA_FIELD GCS_field;
	double cos_psi;
	double sin_psi;
	double theta = GCS_angle.theta * PI/180.0;
	double phi = GCS_angle.phi * PI / 180.0;

	if ((pow(cos(beta)*cos(gamma)*cos(theta) + (sin(beta)*cos(gamma)*cos(phi - alpha) - sin(gamma)*sin(phi - alpha))*sin(theta), 2) == 1)||(pow(cos(beta)*cos(gamma)*cos(theta) + (sin(beta)*cos(gamma)*cos(phi - alpha) - sin(gamma)*sin(phi - alpha))*sin(theta), 2) == 1))
	{
		theta+=0.0000001;
	}
	
	cos_psi = (cos(beta)*cos(gamma)*sin(theta) - (sin(beta)*cos(gamma)*cos(phi - alpha) - sin(gamma)*sin(phi - alpha))*cos(theta)) / sqrt(1.0 - pow(cos(beta)*cos(gamma)*cos(theta) + (sin(beta)*cos(gamma)*cos(phi - alpha) - sin(gamma)*sin(phi - alpha))*sin(theta), 2));
	sin_psi = (sin(beta)*cos(gamma)*sin(phi - alpha) + sin(gamma)*cos(phi - alpha)) / sqrt(1.0 - pow(cos(beta)*cos(gamma)*cos(theta) + (sin(beta)*cos(gamma)*cos(phi - alpha) - sin(gamma)*sin(phi - alpha))*sin(theta), 2));;
	

	GCS_field.theta = cos_psi * LCS_field.theta - sin_psi * LCS_field.phi;
	GCS_field.phi = sin_psi * LCS_field.theta + cos_psi * LCS_field.phi;

	return GCS_field;

}


