#include "ue.h"
#include <stdio.h>
#include <iostream>
#include <fstream>
#include "const.h"
#include "struct.h"
#include "paramSimul.h"
#include "paramFastFading.h"
#include "evaluationConfigurerations.h"
#include "objectBSSectorUe.h"
#include "channel.h"
using namespace std;

/*===================================================================
FUNCTION: UE::UE()

AUTHOR: SJ

DESCRIPTION: UE Object Constructor

NOTES: Currently, nothing is done when a UE is instantiated

===================================================================*/
UE::UE()
{
	adj_bs = new int[num_compute_coef];
	
}

/*===================================================================
FUNCTION: UE::~UE()


AUTHOR: SJ

DESCRIPTION: UE Object Destructor

NOTES: Currently, nothing is done

===================================================================*/
UE::~UE()
{	
	delete[] adj_bs;

}


/*===================================================================
FUNCTION: UE::Configure( int ue_idx )

AUTHOR: JH

DESCRIPTION: UE object gets to know the index of itself

NOTES:

===================================================================*/
void UE::Configure(int ue_idx)
{

	Pathloss_car = 0;
	Pathloss_tw = 0;
	Set_UE_speed_indoor_outdoor_state();
	Set_h_UT();
	_ue_idx = ue_idx;
	uplinkThroghput = 0;
	HARQeSINR = 0;

	alpha = Rand.u()*2.0*PI - PI;
	beta = 0;
	gamma = 0;
}


void UE::Allocate_memory()
{


}


void UE::Set_UE_speed_indoor_outdoor_state()
{
	if (test_environment <3) //Indoor
	{
		speed = indoor_ue_speed * 1000. / 3600.;
	}
	else if (test_environment<9) // UMa, UMi
	{
		if (Rand.u() < indoor_ue_portion) // if (Rand.u() < indoor_ue_portion)  indoor outdoor 
		{
			ue_indoor_outdoor_state = INDOOR;
			
			speed = indoor_ue_speed * 1000. / 3600.;
		}
		else
		{
			ue_indoor_outdoor_state = IN_CAR;

			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}		
	}
	else //Rural
	{
		if (Rand.u() < indoor_ue_portion)
		{
			ue_indoor_outdoor_state = INDOOR;
			
			speed = indoor_ue_speed * 1000. / 3600.;
		}
		else if(Rand.u()< outdoor_pedestrian_portion/(1- indoor_ue_portion))
		{
			ue_indoor_outdoor_state = PEDESTRIAN;
			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}
		else
		{
			ue_indoor_outdoor_state = IN_CAR;
			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}
	}

	double theta_v = PI/2.;
	double phi_v = 2. * PI*Rand.u();
	
	
	velocity.x = speed * sin(theta_v)*cos(phi_v);
	velocity.y = speed * sin(theta_v)*sin(phi_v);
	velocity.z = speed * cos(theta_v);

	
	if (test_environment >= 3)
	{
		if (ue_indoor_outdoor_state == INDOOR)
		{
			Set_2D_distance_indoor();

			if (Rand.u() < high_loss_type_portion)
			{
				O2I_penetration_loss_model = High_loss_model;
			}
			else
			{
				O2I_penetration_loss_model = Low_loss_model;
			}
		}
		else if (ue_indoor_outdoor_state == IN_CAR)
		{
			Set_car_penet_loss();
		}
	}
		

	if ((ue_indoor_outdoor_state==INDOOR) && (test_environment>=3))
	{
		Calculate_O2I_penet_loss(_2D_distance_indoor);
	}
	
}


/*===================================================================
FUNCTION: UE::Set_2D_distance_indoor( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_2D_distance_indoor

NOTES: m.eval2020 page 44,45

===================================================================*/
void UE::Set_2D_distance_indoor()
{	
	if (pathloss_model == UMa_A || pathloss_model == UMi_A)
	{
		if (carrier_freq < 6000000000)
		{
			_2D_distance_indoor = Rand.u() * 25;
		}
		else
		{
			_2D_distance_indoor = MIN(Rand.u() * 25, Rand.u() * 25);
		}

	}
	else if (pathloss_model==RMa_A)
	{
		if (carrier_freq < 6000000000)
		{
			_2D_distance_indoor = Rand.u() * 10;
		}
		else
		{
			_2D_distance_indoor = MIN(Rand.u() * 10, Rand.u() * 10);
		}
	}
	else if (pathloss_model == UMa_B || pathloss_model == UMi_B)
	{
		_2D_distance_indoor = MIN(Rand.u() * 25, Rand.u() * 25);
	}
	else if (pathloss_model == RMa_B)
	{
		_2D_distance_indoor = MIN(Rand.u() * 10, Rand.u() * 10);
	}


}



/*===================================================================
FUNCTION: UE::Set_O2I_penet_loss( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_O2I_penet_loss

NOTES: m.eval2020 page 44,45

===================================================================*/

void UE::Calculate_O2I_penet_loss(double distance_indoor)
{	

	if (ue_indoor_outdoor_state == INDOOR)
	{
		if ((pathloss_model == UMa_A || pathloss_model == UMi_A) && carrier_freq < 6000000000)
		{
			Pathloss_tw = 20 + 0.5*distance_indoor;
		}
		else if (pathloss_model == RMa_A && carrier_freq < 6000000000)
		{
			Pathloss_tw = 10 + 0.5*distance_indoor;
		}
		else
		{
			if (O2I_penetration_loss_model == High_loss_model)
			{

				Pathloss_tw = High_loss_model_PL + 0.5*distance_indoor +High_loss_model_std* Rand.n();
			}
			else
			{
				Pathloss_tw = Low_loss_model_PL + 0.5*distance_indoor +Low_loss_model_std*Rand.n();
	
			}
		}
	}
	else
	{
		Pathloss_tw = 0;
	}

}


/*===================================================================
FUNCTION: UE::Set_car_penet_loss( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_car_penet_loss

NOTES: m.eval2020 page 45

===================================================================*/

void UE::Set_car_penet_loss()
{
	Pathloss_car = car_penet_loss_mean + car_penet_loss_std*Rand.n();
}


/*===================================================================
FUNCTION: UE::Set_h_UT( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_h_UT

NOTES: m.eval2020 page 19 Table 8-X1 UE antenna height

===================================================================*/
void UE::Set_h_UT()
{
	double n_fl;
	double N_fl;


	if (test_environment >= 3 && test_environment < 7)
	{
		if (ue_indoor_outdoor_state == INDOOR)
		{
			N_fl = floor(5.*Rand.u() + 4);
			n_fl = floor(N_fl*Rand.u() + 1);
			h_UT = 3. * (n_fl - 1) + 1.5;
			floor_idx = int(n_fl) - 1;

		}
		else
		{
			h_UT = 1.5;
			floor_idx = 0;
		}
	}
	else
	{
		floor_idx = 0;
		h_UT = 1.5;
	}
}


/*===================================================================
FUNCTION: Drop_ue_in_cell()

AUTHOR: G.J

DESCRIPTION: Drop ues

NOTES: UE  sector Ȥ cell ȿ ߸.

OUTPUT : location_x, location_y

===================================================================*/
void UE::Drop_ue_in_site()
{

	if (test_environment == Dense_Urban_eMBB_C_Micro)
	{
		do
		{
			location_in_site.x = micro_cell_radius * Rand.u();
			location_in_site.y = micro_cell_radius * Rand.u();

		} while ((sqrt(location_in_site.x*location_in_site.x + location_in_site.y*location_in_site.y) < secter_to_ue_min_distance) || (sqrt(location_in_site.x*location_in_site.x + location_in_site.y*location_in_site.y) > micro_cell_radius));

		double x = Rand.u();
		if (x < 0.25)
		{
			location_in_site.y = -location_in_site.y;
		}
		else if (x < 0.5)
		{
			location_in_site.x = -location_in_site.x;
			location_in_site.y = -location_in_site.y;
		}
		else if (x < 0.75)
		{
			location_in_site.x = -location_in_site.x;
		}
	}
	else
	{

		double x;
		double y;
		
		if ((_ue_idx/num_user_per_sector) % 3 == 0)
		{
			
			do
			{
				x = inter_site_distance / sqrt(3.)* Rand.u();
				y = inter_site_distance * (1. / 2. + 1. / 4.)* Rand.u() - 1. / 4.*inter_site_distance;
			} while ((y > -sqrt(3.)*(x - 1. / sqrt(3.)*inter_site_distance)) || (y < sqrt(3.)*(x - 1. / sqrt(3.)*inter_site_distance)) || (y < -1. / sqrt(3.)*(x)) || (sqrt(x*x + y*y) < secter_to_ue_min_distance));
		}
		else if ((_ue_idx / num_user_per_sector) % 3 == 1)
		{
			
			do
			{
				x = - inter_site_distance / sqrt(3.)* Rand.u();
				y = inter_site_distance * (1. / 2. + 1. / 4.)* Rand.u() - 1. / 4.*inter_site_distance;
			} while ((y > sqrt(3.)*(x + 1. / sqrt(3.)*inter_site_distance)) || (y < -sqrt(3.)*(x + 1. / sqrt(3.)*inter_site_distance)) || (y < 1. / sqrt(3.)*(x)) || (sqrt(x*x + y * y) < secter_to_ue_min_distance));
		}
		else
		{
			
			do
			{
				x = inter_site_distance * (sqrt(3.) / 2)* Rand.u() - inter_site_distance* (sqrt(3.) / 4);
				y = -inter_site_distance / 2.* Rand.u();

			} while ( (y < -sqrt(3.)*(x + 1. / sqrt(3.)*inter_site_distance)) || (y < sqrt(3.)*(x - 1. / sqrt(3.)*inter_site_distance)) || (y > 1. / sqrt(3.)*(x)) || (y > -1. / sqrt(3.)*(x)) || (sqrt(x*x + y * y) < secter_to_ue_min_distance));
		}
		
		location_in_site.x = x;
		location_in_site.y = y;
		//cout << pos_in_site.y << endl;
		
		/*
		bool a = 0;

		if (a == 0)
		{
			do
			{
				pos_in_cell.x = inter_site_distance / 2.* Rand.u();
				pos_in_cell.y = inter_site_distance / sqrt(3.)* Rand.u();
			} while (pos_in_cell.y > (-pos_in_cell.x + inter_site_distance) / sqrt(3.) || sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance);
		}
		else
		{
			do
			{
				pos_in_cell.y = inter_site_distance / 2.* Rand.u();
				pos_in_cell.x = inter_site_distance / sqrt(3.)* Rand.u();
			} while (pos_in_cell.x > (-pos_in_cell.y + inter_site_distance) / sqrt(3.)	|| sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance);

		}

		double x = Rand.u();

		if (x < 0.25)
		{
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.5)
		{
			pos_in_cell.x = -pos_in_cell.x;
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.75)
		{
			pos_in_cell.x = -pos_in_cell.x;
		}
		*/
	}
}

/*===================================================================
FUNCTION: Determine_LOS()

AUTHOR: G.J

DESCRIPTION: Determine_LOS

NOTES: m.2020 eval page 46

OUTPUT : ue[ue_idx].los[sec_idx]

===================================================================*/
void UE::Determine_LOS(int bs_idx)
{
	double d_2D =  channel[bs_idx][_ue_idx]._2D_distance;
	double d_in=0;

	if ((test_environment>=3) && (ue_indoor_outdoor_state == INDOOR))
	{
		if (d_2D < _2D_distance_indoor)
		{			
			d_in = d_2D;
		}
		else
		{
			d_in = _2D_distance_indoor;
		}
	}
	else
	{
		d_in = 0;
	}
	double d_2D_out = d_2D - d_in;
	double P_LOS;



	if (test_environment <3)
	{
		if (d_2D < 5)
		{
			P_LOS = 1.0;
		}
		else if(d_2D < 49)
		{
			P_LOS = exp(-(d_2D - 5.) / 70.8);
		}
		else
		{
			P_LOS = exp(-(d_2D - 49.) / 211.7)*0.54;
		}
	}
	else if (test_environment<6)
	{
		if (d_2D <= 18.0)
		{
			P_LOS = 1.0;
		}
		else
		{
			double C;
				
			if (h_UT <= 13.0)
			{
				C = 0;
			}
			else
			{
				C = pow((h_UT - 13.0) / 10.0, 1.5);
			}

			if (ue_indoor_outdoor_state== INDOOR)
			{
				P_LOS = (18.0 / d_2D_out + exp(-d_2D_out / 63)*(1 - 18 / d_2D_out)) * (1 + C*5.0 / 4.0*pow(d_2D_out / 100.0, 3)*exp(-d_2D_out / 150.0));
			}
			else
			{
				P_LOS = (18.0 / d_2D + exp(-d_2D / 63)*(1 - 18 / d_2D)) * (1 + C*5.0 / 4.0*pow(d_2D / 100.0, 3)*exp(-d_2D / 150.0));
			}
		}
	}
	else if (test_environment == 6)
	{
		if (d_2D <= 18.0)
		{
			P_LOS = 1.0;
		}
		else
		{
			if (ue_indoor_outdoor_state == INDOOR)
			{
				P_LOS = (18.0 / d_2D_out + exp(-d_2D_out / 36)*(1 - 18 / d_2D_out));
			}
			else
			{
				P_LOS = (18.0 / d_2D + exp(-d_2D / 63)*(1 - 18 / d_2D));
			}
		}
	}
	else
	{
		if (d_2D <= 10.)
		{
			P_LOS = 1.0;
		}
		else
		{
			if (ue_indoor_outdoor_state == INDOOR)
			{
				P_LOS = exp(-(d_2D_out - 10)/1000);
			}
			else
			{
				P_LOS = exp(-(d_2D - 10) / 1000);
			}
		}

	}
	

	if ( Rand.u() < P_LOS)
	{
		channel[bs_idx][_ue_idx].propagation_condition = LOS;
		channel[bs_idx][_ue_idx].fast_fading_condition = LOS;
	}
	else
	{
		channel[bs_idx][_ue_idx].propagation_condition = NLOS;
		channel[bs_idx][_ue_idx].fast_fading_condition = NLOS;
		
	}
	
	
	if ((test_environment>=3)&&(ue_indoor_outdoor_state == INDOOR))
	{
		channel[bs_idx][_ue_idx].fast_fading_condition = OUT2IN;
	}

	
	 //ue_indoor_outdoor_state = IN_CAR;

	//LOS/NLOS
	//channel[bs_idx][_ue_idx].propagation_condition = LOS;
	//channel[bs_idx][_ue_idx].fast_fading_condition = LOS;
	

}


/*===================================================================
FUNCTION: void UE::Set_breakpoint_distance(int sec_idx)

AUTHOR: G.J

DESCRIPTION: Set_breakpoint_distance

NOTES: 

OUTPUT : ue[ue_idx].break_point_distance[sec_idx]

===================================================================*/
void UE::Set_breakpoint_distance(int bs_idx)
{
	double d_2D = channel[bs_idx][_ue_idx]._2D_distance;;

	if (test_environment<7)// not rural
	{
		double C;
		double g;
		double h_E;


		if (d_2D <= 18.)
		{
			g = 0.;
		}
		else
		{
			g = 5. / 4. * pow(d_2D / 100., 3.)*exp(-d_2D / 150.);
		}

		if (h_UT < 13.)
		{
			C = 0;
		}
		else
		{
			C = pow((h_UT - 13.) / 10., 1.5)*g;
		}


		if (Rand.u() < 1 / (1 + C))
		{
			h_E = 1.;
		}
		else
		{
			h_E = floor((h_UT - 1.5 - 9.) / 3. * Rand.u()) * 3. + 12.;
		}

		channel[bs_idx][_ue_idx].break_point_distance = 4. * (h_BS - h_E) * (h_UT - h_E) * carrier_freq / SPEED_LIGHT;

	}
	else
	{		
		channel[bs_idx][_ue_idx].break_point_distance = 2. * PI * h_BS * h_UT * carrier_freq / SPEED_LIGHT;
	}
}


/*===================================================================
FUNCTION: void UE::Compute_PathLoss(int sec_idx)

AUTHOR: G.J

DESCRIPTION: Compute_PathLoss

NOTES:

OUTPUT : ue[ue_idx].pathloss[bs_idx]

===================================================================*/
void UE::Compute_PathLoss(int bs_idx)
{
	double d_2D = channel[bs_idx][_ue_idx]._2D_distance;
	double d_in = 0;


	double d_3D = pow(pow(d_2D, 2.) + pow(h_BS - h_UT, 2.), 0.5);
	double d_BP = channel[bs_idx][_ue_idx].break_point_distance;
	double fc = carrier_freq / 1000000000.;
	
	channel[bs_idx][_ue_idx]._3D_distance = d_3D;
	
	

	if (channel[bs_idx][_ue_idx].propagation_condition == LOS)
	{
		if (pathloss_model == InH_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.8 + 16.9*log10(d_3D) + 20 * log10(fc);

					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_A " << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 17.3*log10(d_3D) + 20 * log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_A " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc, LOS Compute_PathLoss pathloss_model InH_A" << endl;
				getchar();
			}
		}
		else if (pathloss_model == InH_B)
		{
			if (fc >= 0.5 &&fc <= 100)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 17.3*log10(d_3D) + 20 * log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc, LOS Compute_PathLoss pathloss_model InH_B" << endl;
				getchar();
			}

		}
		else if (pathloss_model == UMa_A)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					
					cout << "check d_2D, LOS Compute_PathLoss UMa_A " << endl;
					getchar();
				}

			
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMa_A" << endl;
				getchar();
			}
			
		}
		else if (pathloss_model == UMa_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMa_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMa_B" << endl;
				getchar();
			}
		}
		else if (pathloss_model == UMi_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 22.*log10(d_3D) + 28.0 + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_A " << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_A " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMi_A" << endl;
				getchar();
			}

		}
		else if (pathloss_model == UMi_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMi_B" << endl;
				getchar();
			}
		}
		else if (pathloss_model == RMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D;
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					channel[bs_idx][_ue_idx].pathloss_dB = PL1 + 40 * log10(d_3D / d_BP);
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss RMa_A " << endl;
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model RMa_A" << endl;
				getchar();
			}

		}
		else if (pathloss_model == RMa_B)
		{
			if (fc >= 0.5 && fc <= 30)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D;
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					channel[bs_idx][_ue_idx].pathloss_dB = PL1 + 40 * log10(d_3D / d_BP) ;
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss RMa_B " << endl;
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model RMa_B" << endl;
				getchar();
			}
		}

		
	}
	else
	{
		double PL_LOS;
		double PL_NLOS;

		if (pathloss_model == InH_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (d_2D <= 150)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 43.3*log10(d_3D) + 11.5 + 20.*log10(fc);
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
					cout << "bs_idx : "<< bs_idx << endl;
					cout << "ue_idx : "<<_ue_idx << endl;
					cout << channel[bs_idx][_ue_idx]._2D_distance << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (d_2D <= 150)
				{
					PL_LOS = 32.4 + 17.3*log10(d_3D) + 20.0*log10(fc);
					PL_NLOS = 17.3 + 38.3*log10(d_3D) + 24.9*log10(fc);
					channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
					if(PL_LOS>PL_NLOS)
					{
						channel[bs_idx][_ue_idx].sigma_SF = 3;
					}
					else
					{
						channel[bs_idx][_ue_idx].sigma_SF = 8.03;
					}
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss InH_A" << endl;
			}
		}
		else if (pathloss_model == InH_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (d_2D <= 150.)
				{
					PL_LOS = 32.4 + 17.3*log10(d_3D) + 20.0*log10(fc);
					PL_NLOS = 17.3 + 38.3*log10(d_3D) + 24.9*log10(fc);
					channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
					if (PL_LOS>PL_NLOS)
					{
						channel[bs_idx][_ue_idx].sigma_SF = 3;
					}
					else
					{
						channel[bs_idx][_ue_idx].sigma_SF = 8.03;
					}
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss InH_A" << endl;
			}
		}
		else if (pathloss_model == UMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMa_A " << endl;
				}

				if (10. <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3)
						+ 20.*log10(fc) - (3.2*pow(log10(17.625), 2) - 4.97) - 0.6*(h_UT - 1.5);
				}
				else
				{
					
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				

				
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);

				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}

			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 13.54 + 39.08*log10(d_3D) + 20 * log10(fc) - 0.6 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMa_A" << endl;
			}

		}
		else if (pathloss_model == UMa_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_B " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 13.54 + 39.08*log10(d_3D) + 20 * log10(fc) - 0.6 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_B " << endl;
				}
				
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
			/*	cout << h_BS << endl;
				cout << h_UT << endl;
				cout << "d_BP :  "<< d_BP << endl;
				cout << "PL_LOS : " << PL_LOS << endl;
				cout << "PL_NLOS: " << PL_NLOS << endl;
				cout << -channel[bs_idx][_ue_idx].pathloss_dB << endl;
				getchar();*/

				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 6;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMa_B" << endl;
			}



		}
		else if (pathloss_model == UMi_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc) ;
				}
				else if (d_BP < d_2D	&& d_2D < 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.) // !!!!!! Table 3.3 10m<d_2d<2000m ,
				{
					PL_NLOS = 36.7*log10(d_3D) + 22.7 + 26 * log10(fc) - 0.3*(h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB= MAX(PL_LOS, PL_NLOS);

				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 3;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}

			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP < d_2D	&& d_2D < 5000.)
				{
					PL_LOS = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}

				if (secter_to_ue_min_distance < d_2D && d_2D < 5000.)   ///!! table3-3 shadowing check in 38.901 7.82
				{
					PL_NLOS = 35.3*log10(d_3D) + 22.4 + 21.3 * log10(fc) - 0.3 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 7.82;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMi_A" << endl;
			}
		}
		else if (pathloss_model == UMi_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_B " << endl;
					getchar();
				}

				if (10. < d_2D && d_2D < 5000.)  ///!! table3-3 shadowing check
				{
					PL_NLOS = 35.3*log10(d_3D) + 22.4 + 21.3* log10(fc) - 0.3 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}
				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					channel[bs_idx][_ue_idx].sigma_SF = 4;
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 7.82;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMi_A" << endl;
			}
		}
		else if (pathloss_model == RMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= 21000.)
				{
					channel[bs_idx][_ue_idx].pathloss_dB = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3) + 20 * log10(fc) - (3.2*pow(log10(11.75*h_UT), 2) - 4.97);
					channel[bs_idx][_ue_idx].sigma_SF = 8;
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss RMa_A" << endl;
			}
		}
		else if (pathloss_model == RMa_B)
		{
			if (fc >= 0.5 && fc <= 30)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D ;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					PL_LOS = PL1 + 40 * log10(d_3D / d_BP);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_B " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 21000.)
				{
					PL_NLOS = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3) + 20 * log10(fc) - (3.2*pow(log10(11.75*h_UT), 2) - 4.97);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_B " << endl;
				}

				channel[bs_idx][_ue_idx].pathloss_dB = MAX(PL_LOS, PL_NLOS);
				
				if (PL_LOS>PL_NLOS)
				{
					if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
					{
						channel[bs_idx][_ue_idx].sigma_SF = 4;
					}
					else
					{
						channel[bs_idx][_ue_idx].sigma_SF = 6;
					}
					
				}
				else
				{
					channel[bs_idx][_ue_idx].sigma_SF = 8;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss RMa_A" << endl;
			}
		}


	}

	if (channel[bs_idx][_ue_idx].fast_fading_condition == OUT2IN)
	{
		if (pathloss_model == RMa_A || pathloss_model == RMa_B)
		{
			channel[bs_idx][_ue_idx].sigma_SF = 8;
		}
		else
		{
			channel[bs_idx][_ue_idx].sigma_SF = 7;
		}
		
	}

	channel[bs_idx][_ue_idx].pathloss_dB += Pathloss_tw+Pathloss_car;

	channel[bs_idx][_ue_idx].pathloss = dB2linear(-channel[bs_idx][_ue_idx].pathloss_dB);

}


/*
void UE::Update_Geometry(void)
{
	num_of_rbs = 0;
	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{
		SCHEDULE_DECISION decision_of_interest = ppSchedule_Map[sector_in_control][rb_idx];
		if (decision_of_interest.ue_selected == _ue_idx)
		{
			rb_indices.push_back(rb_idx);
			num_of_rbs++;
		}
	}

	double denominator;
	double interference;
	double thermal;
		int *ue_same_rs;
	ue_same_rs = new int[num_sector];

	thermal = dBm2linear(thermal_nosie_level);// *UE_BANDWIDTH;


	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{
		denominator = 0.;
		interference = 0.;
		for (int sec_idx = 0; sec_idx < num_sector; sec_idx++)
		{
			ue_same_rs[sec_idx] = ppSchedule_Map[sec_idx][rb_idx].ue_selected;
			if (sector_in_control != sec_idx && ue_same_rs[sec_idx] >= 0)
			{
				interference += dBm2linear(ue[ue_same_rs[sec_idx]].LS_gain[sector_in_control]+ ue[ue_same_rs[sec_idx]].Update_Power_Control(rb_idx));
			}
		}

		denominator = thermal + interference;
		geometry[rb_idx] = linear2dB(dBm2linear(LS_gain[sector_in_control] + Update_Power_Control(rb_idx)) / (denominator));
	}
	
}
*/

/*===================================================================
FUNCTION: UE::Update_Power_Control( void )

AUTHOR: JH

DESCRIPTION: return the total throughput over the frequency

NOTES:

===================================================================*/
double UE::Update_Power_Control(int rb_idx)
{
	double power_per_rb;
	if (num_of_rbs == 0)
	{
		return 0.0;
	}
	else
	{
		double P_max = 10 * log10(250);
		int P0 = -106;
		int M_pusch = num_of_rbs;
		double Ks = 1.25;
		int N_info = info_bits;
		int N_re = num_of_rbs * 12 * 14;
		double MPR = (double)N_info / (double)N_re;
		double delta_TF = 0.;// 10*log10( pow( 2. , MPR*Ks) - 1);
		double f_of_i = 0.;
		double alpha = 0.8;
		double Total_PL = -channel[sector_in_control/3][_ue_idx].LS_gain;  //  ʿ-channel[sector_in_control][_ue_idx].LS_gain .

		power_per_rb = MIN(P_max - 10 * log10(M_pusch), P0 + alpha * Total_PL + delta_TF + f_of_i);

		return (power_per_rb); // (num_tx_antenna_element);
		//return 0.0;

	}
}

void UE::Receive_UL()
{
	double instantThroughput = 0;
	Find_Allocated_Rbs_Mcs();

	if (rb_indices.size() > 0)
	{
		esinr =Compute_Effective_Snr();
		spectralEfficiency = Get_SpectralEfficiency(esinr, mcs);

		if (FER(esinr, mcs)< Rand.u())
		{
			instantThroughput = spectralEfficiency * simulation_bandwidth / 10 /1000000 * 50 * (12 * 11 - 12 - 16) * 1000;
			HARQeSINR = 0;
		}
		else
		{
			instantThroughput = 0;
			HARQeSINR = HARQeSINR + esinr;
		}
	}
	uplinkThroghput = uplinkThroghput  + instantThroughput ;


}


void UE::Find_Allocated_Rbs_Mcs(void)
{
	SCHEDULE_DECISION decision_of_interest;

	rb_indices.clear();
	for (int freq_idx = 0; freq_idx < num_rb; freq_idx++)
	{
		decision_of_interest = ppSchedule_Map[sector_in_control][freq_idx];
		if (decision_of_interest.ue_selected == _ue_idx)
		{
			rb_indices.push_back(freq_idx);
			if (rb_indices.size() == 1)
			{
				mcs = decision_of_interest.mcs_selected;
			}
			else
			{
				if (mcs != decision_of_interest.mcs_selected)
				{
					cerr << "Something Wrong in the Scheduler" << endl;
				}
			}
		}
	}
}


double UE::Compute_Effective_Snr()
{
	double a=0;
	return a;
}

double UE::FER(double esinr, int mcs)
{
	double Frame_Error_Rate = 1;
	double ESINR_L = esinr; // linear value
	switch (mcs)
	{
	case -1:
		Frame_Error_Rate = 0.0;
		break;

	case 0:
		if (ESINR_L < pow(10.0, -7.737 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(69.109 - 410.4 * ESINR_L);
		break;
	case 1:
		if (ESINR_L < pow(10.0, -5.852 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(70.072 - 269.63 * ESINR_L);

		break;

	case 2:
		if (ESINR_L < pow(10.0, -3.737 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(69.19 - 163.42 * ESINR_L);

		break;

	case 3:
		if (ESINR_L < pow(10.0, -1.718 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(70.491 - 104.69 * ESINR_L);

		break;

	case 4:
		if (ESINR_L < pow(10.0, 0.3206 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(77.142 - 71.651* ESINR_L);

		break;

	case 5:
		if (ESINR_L < pow(10.0, 2.14 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(47.496 - 29.018 * ESINR_L);

		break;

	case 6:
		if (ESINR_L < pow(10.0, 4.096 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(45.163 - 17.585 * ESINR_L);

		break;

	case 7:
		if (ESINR_L < pow(10.0, 6.052 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(45.936 - 11.402 * ESINR_L);

		break;

	case 8:
		if (ESINR_L < pow(10.0, 8.03 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(44.625 - 7.0239 * ESINR_L);

		break;

	case 9:
		if (ESINR_L < pow(10.0, 10.03 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(46.766 - 4.6398 * ESINR_L);

		break;

	case 10:
		if (ESINR_L < pow(10.0, 11.82 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(41.834 - 2.7525 * ESINR_L);

		break;
	case 11:
		if (ESINR_L < pow(10.0, 13.65 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(33.542 - 1.4485 * ESINR_L);

		break;
	case 12:
		if (ESINR_L < pow(10.0, 15.69 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(29.392 - 0.79247 * ESINR_L);

		break;
	case 13:
		if (ESINR_L < pow(10.0, 17.5 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(29.525 - 0.52504 * ESINR_L);

		break;
	case 14:
		if (ESINR_L < pow(10.0, 18.98 / 10.0))
			Frame_Error_Rate = 1.0;
		else
			Frame_Error_Rate = exp(8.154 - 0.10315 * ESINR_L);

		break;
	}
	return Frame_Error_Rate;
}

double UE::Get_SpectralEfficiency(double SINR, int MCS)
{
	if (SINR < -7.8474) { MCS = -1; return 0.0; }
	else if (SINR >= -7.8474&&SINR < -6.2369) { MCS = 0; return  0.1523; }
	else if (SINR >= -6.2369&&SINR < -4.3591) { MCS = 1; return 0.2344; }
	else if (SINR >= -4.3591&&SINR < -1.9319) { MCS = 2; return 0.3770; }
	else if (SINR >= -1.9319&&SINR < 0.1509) { MCS = 3; return 0.6016; }
	else if (SINR >= 0.1509&&SINR < 1.9976) { MCS = 4; return 0.8770; }
	else if (SINR >= 1.9976&&SINR < 4.7278) { MCS = 5; return 1.1758; }
	else if (SINR >= 4.7278&&SINR < 6.2231) { MCS = 6; return 1.4766; }
	else if (SINR >= 6.2231&&SINR < 8.0591) { MCS = 7; return 1.9141; }
	else if (SINR >= 8.0591&&SINR < 9.8585) { MCS = 8; return 2.4063; }
	else if (SINR >= 9.8585&&SINR < 11.8432) { MCS = 9; return 2.7305; }
	else if (SINR >= 11.8432&&SINR < 13.4893) { MCS = 10; return 3.3223; }
	else if (SINR >= 13.4893&&SINR < 15.3598) { MCS = 11; return 3.9023; }
	else if (SINR >= 15.3598&&SINR < 17.4435) { MCS = 12; return 4.5234; }
	else if (SINR >= 17.4435&&SINR < 19.2155) { MCS = 13; return 5.1152; }
	else if (SINR >= 19.2155) { MCS = 14; return 5.5547; }
	return -1.0;
}

/*

void UE::Update_Channel(void)
{
	complex <double> _dft_of_channel[num_rbs_in_1ms];
	int num_of_channel = num_received_antenna * num_transmit_antenna;

	if (_ch_gamma == NULL)
	{
		_ch_gamma = new double **[num_received_antenna];
		_gamma = new double[num_rbs_in_1ms];
		for (int r_idx = 0; r_idx < num_received_antenna; r_idx++)
		{
			_ch_gamma[r_idx] = new double *[num_transmit_antenna];
			for (int t_idx = 0; t_idx < num_transmit_antenna; t_idx++)
			{
				_ch_gamma[r_idx][t_idx] = new double[num_rbs_in_1ms];
			}
		}
	}

	for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
	{
		_gamma[rbs_idx] = 0.;
	}

	if (_p_of_k == NULL)
	{
		_p_of_k = new double **[num_received_antenna];
		for (int r_idx = 0; r_idx < num_received_antenna; r_idx++)
		{
			_p_of_k[r_idx] = new double *[num_transmit_antenna];
			for (int t_idx = 0; t_idx < num_transmit_antenna; t_idx++)
			{
				_p_of_k[r_idx][t_idx] = new double[num_rbs_in_1ms];
			}
		}
	}

	// compute dft of channel per resource block
	for (int r_idx = 0; r_idx < num_received_antenna; r_idx++)
	{
		for (int t_idx = 0; t_idx < num_transmit_antenna; t_idx++)
		{
			double sum_for_normalize = 0.0;
			int pointer = (int(DFT_SIZE) - (num_rbs_in_1ms*num_freq_per_rbs)) / 2;

			int indoor = (int)(Rand.u() + 0.5);

			for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
			{
				int freq_idx = (pointer + (num_freq_per_rbs / 2) + (rbs_idx*num_freq_per_rbs)) * 15000;
				_dft_of_channel[rbs_idx].real() = 0.;
				_dft_of_channel[rbs_idx].imag() = 0.;

				for (int path_idx = 0; path_idx < num_path[_self_idx][0]; path_idx++)
				{
					complex <double> dft_exp(0.0, -(2 * PI*freq_idx*(delay[_self_idx][0][path_idx])));

					_dft_of_channel[rbs_idx] += Channel_coef[_self_idx][0][path_idx][r_idx][t_idx] * exp(dft_exp);
				}
				_p_of_k[r_idx][t_idx][rbs_idx] = pow(abs(_dft_of_channel[rbs_idx]), 2.);
				sum_for_normalize += _p_of_k[r_idx][t_idx][rbs_idx];
			}
			sum_for_normalize = sum_for_normalize / (double)num_rbs_in_1ms;
			// compute snr per per resource block
			for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
			{
				_p_of_k[r_idx][t_idx][rbs_idx] = _p_of_k[r_idx][t_idx][rbs_idx] / sum_for_normalize;
				_ch_gamma[r_idx][t_idx][rbs_idx] = _p_of_k[r_idx][t_idx][rbs_idx] * dB2linear(_geometry[rbs_idx]) *
					(DFT_SIZE / (DFT_SIZE + double(CYCLIC_PREFIX_LEN))) *
					(0.8* 8400.) / 7200.;
			}
		}
	}

	for (int rbs_idx = 0; rbs_idx < num_rbs_in_1ms; rbs_idx++)
	{
		for (int r_idx = 0; r_idx < num_received_antenna; r_idx++)
		{
			for (int t_idx = 0; t_idx < num_transmit_antenna; t_idx++)
			{
				_gamma[rbs_idx] += _ch_gamma[r_idx][t_idx][rbs_idx];
			}
		}
		_gamma[rbs_idx] = linear2dB(_gamma[rbs_idx]);
	}

	// decide mcs type
	for (int freq_idx = 0; freq_idx < num_rbs_in_1ms; freq_idx++)
	{
		for (int threshold_idx = 0; threshold_idx < NUM_CQI_MCS_TYPES - 1; threshold_idx++)
		{
			if (_gamma[freq_idx] < Threshold[0])
			{
				_channel_quality_freq[freq_idx] = 0;
			}
			else if (_gamma[freq_idx] >= Threshold[threshold_idx] && _gamma[freq_idx]  < Threshold[threshold_idx + 1])
			{
				_channel_quality_freq[freq_idx] = threshold_idx;
			}
			else if (_gamma[freq_idx] > Threshold[NUM_CQI_MCS_TYPES - 1])
			{
				_channel_quality_freq[freq_idx] = NUM_CQI_MCS_TYPES - 1;
			}
		}
	}

	if (drop_idx == ue_distribution - 1)
	{
		Get_Gamma_Distribution();
	}

	for (int rb_idx = 0; rb_idx < num_rbs_in_1ms; rb_idx++)
	{
		_gamma[rb_idx] = dB2linear(_gamma[rb_idx]);
	}
}



void UE::Receive_UL(void)
{
	Find_Allocated_Rbs_Mcs();

	// if UE has at least more than 1 resource block
	if (rb_indices.size() > 0)
	{
		Compute_Num_Bin_Symbol();
		Compute_Code_Rate();
		//Compute_Beta();
		Compute_Effective_Snr();

		Get_SINR(_self_idx, linear2dB(effective_snr));
		total_num_tx++;
		if (effective_snr < exp(100000000) && nack == true)
		{

			sum_of_esnr += effective_snr;
			num_of_retx++;
			if (num_of_retx > 3)
			{
				num_of_retx = 0;
				nack = false;
			}

			if (effective_snr < 0)
			{
				cout << sum_of_esnr << endl;
			}
		}

		// Now we can look at the table & Tossing a biased coin
		if (ComputeBLER(_input2table) < Rand.u())
		{
			if (_i_tbs >= 0) { _throughput += _info_bits; }
			nack = false;
			num_of_retx = 0;
			sum_of_esnr = 0;
		}
		else
		{
			nack = true;
			if (num_of_retx == 0)
			{
				error_1st++;
			}
			else if (num_of_retx == 1)
			{
				error_2nd++;
			}
			else if (num_of_retx == 2)
			{
				error_3rd++;
			}
			else if (num_of_retx == 3)
			{
				error_4th++;
			}
		}
	}
}
*/